import numpy as np
import tensorflow as tf
from geostat.kernel import quadstack, quadstack1
from geostat.metric import ed

def vanilla_quadstack(x, sills, ranges):
    """
    `x` has arbitrary shape [...], but must be non-negative.
    `sills` and `ranges` both have shape [K].
    """
    r2 = ranges
    r1 = tf.pad(ranges[:-1], [[1, 0]])
    ex = ed(x)
    ax = tf.abs(ex)
    i1 = tf.cast(ax <= r1, tf.float32) # Indicates x <= r1.
    i2 = tf.cast(ax <= r2, tf.float32) * (1. - i1) # Indicates r1 < x <= r2.

    c1 = 2. / (r1 + r2)
    c2 = 1. / (1. - tf.square(r1/r2))

    y = sills * (i1 * (1. - c1 * ax) + i2 * c2 * tf.square(ax / r2 - 1))
    return tf.reduce_sum(y, -1)

def test_quadstack():

    sills = tf.Variable([0.1, 0.2, 0.5, 0.7], dtype=tf.float32)
    ranges = tf.Variable([1, 2, 5, float('inf')], dtype=tf.float32)
    x = tf.Variable([0.5, 1.5, 2.5, 5.5], dtype=tf.float32)

    with tf.GradientTape(persistent=True) as g:
        g.watch(sills)
        g.watch(ranges)
        g.watch(x)
        y1 = quadstack(x, sills, ranges)
        y2 = vanilla_quadstack(x, sills, ranges)

    print('---')
    print(y1)
    print(y2)
    print('---')
    print(g.jacobian(y1, sills))
    print(g.jacobian(y2, sills))
    print('---')
    print(g.jacobian(y1, ranges))
    print(g.jacobian(y2, ranges))
    print('---')
    print(g.jacobian(y1, x))
    print(g.jacobian(y2, x))



import numpy as np
from geostat import GP, Model, NormalizingFeaturizer, Parameters
import geostat.kernel as krn
import geostat.metric as gm

def test_euclidean():
    # Create 100 random locations in a square centered on the origin.
    locs = np.random.uniform(-1., 1., [1000, 3])

    p = Parameters(zscale=5., range=0.5, sill=1., gamma=1., nugget=1.)

    metric = gm.Euclidean(scale=[1., 1., p.zscale])
    gp = GP(0,
        krn.GammaExponential(range=p.range, sill=p.sill, gamma=p.gamma, metric=metric) +
        krn.SquaredExponential(range=2., sill=p.sill) +
        krn.Noise(nugget=p.nugget))

    # Generating GP.
    model = Model(gp)

    # Generate data.
    vals = model.generate(locs).vals

    # Fit GP.
    model.set(zscale=1., range=1.0, sill=0.5, gamma=0.5, nugget=0.5)
    model.fit(locs, vals, iters=500)

    # Interpolate using GP.
    N = 10
    xx, yy, zz = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy, zz], axis=-1).reshape([-1, 3])

    mean, var = model.predict(locs2)
    mean2, var2 = model.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)

def test_poincare():
    np.random.seed(1235)

    # Create random locations in a square centered on the origin.
    locs = np.random.uniform(-2., 2., [1000, 3])
    locs[:, 2] = 3. - np.exp(locs[:, 2]) # Make z to be 3 or lower.

    p = Parameters(zoff=2.0, alpha=1., zscale=5., range=0.5, sill=1., gamma=1., nugget=1.)

    # Transform z to be positive, and make it the first axis.
    zmax = locs[:, 2].max()
    def xform(x, y, z): return zmax-z, x, y

    metric = gm.Poincare(xform=xform, scale=[p.zscale, 1., 1.], zoff=p.zoff)
    gp = GP(0,
        krn.GammaExponential(range=p.range, sill=p.sill, gamma=p.gamma, metric=metric) +
        krn.SquaredExponential(metric=metric, range=2., sill=p.sill) +
        krn.Noise(nugget=p.nugget))

    # Generating GP.
    model = Model(gp)

    # Generate data.
    vals = model.generate(locs).vals

    # Fit GP.
    model.set(zoff=2.0, zscale=1., range=1.0, sill=0.5, gamma=0.5, nugget=0.5)
    model.fit(locs, vals, iters=500)

    # Interpolate using GP.
    N = 10
    xx, yy, zz = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy, zz], axis=-1).reshape([-1, 3])

    mean, var = model.predict(locs2)
    mean2, var2 = model.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)




import numpy as np
from geostat import GP, Model, Featurizer, NormalizingFeaturizer, Parameters, Trend
import geostat.kernel as krn

def test_coefs1():
    # Create 100 random locations in a square centered on the origin.
    locs1 = np.random.uniform(-1., 1., [100, 2])

    p = Parameters(b=[1.0, 0.5], nugget=0.1)

    # Initialize featurizer of location for trends.
    def trend_terms(x, y): return 1, y
    featurizer = Featurizer(trend_terms)
    gp = GP(Trend(featurizer, beta=p.b),
            krn.SquaredExponential(range=0.5, sill=0.01) + krn.Noise(nugget=p.nugget))

    # Generating GP.
    model = Model(gp)

    # Generate data.
    vals1 = model.generate(locs1).vals

    # Fit GP.
    model.set(b=[1.0, 0.5], nugget=0.1)
    model.fit(locs1, vals1, iters=200)

    # Interpolate using GP.
    N = 2
    xx, yy = np.meshgrid(np.linspace(0, 1, N), np.linspace(0, 1, N))
    locs2 = np.stack([xx, yy], axis=-1).reshape([-1, 2])

    mean, var = model.predict(locs2)

    target_mean = [1., 1., 1.5, 1.5]

    assert np.square(mean - target_mean).mean() < 0.1

def test_coefs2():
    # Create 500 random locations in a square centered on the origin.
    locs1 = np.random.uniform(-1., 1., [500, 2])

    p = Parameters(b1=0.5, b2=0.5, b3=0.25, range=0.5, nugget=1.)

    # Initialize featurizer of location for trends.
    def trend_terms(x, y): return x, y, x*y
    featurizer = NormalizingFeaturizer(trend_terms, locs1)
    gp = GP(Trend(featurizer, beta=[1., p.b1, p.b2, p.b3]),
            krn.SquaredExponential(range=p.range, sill=1.) + krn.Noise(nugget=p.nugget))

    # Generating GP.
    model = Model(gp)

    # Generate data.
    vals1 = model.generate(locs1).vals

    # Fit GP.
    model.set(b1=0., b2=0., b3=0., range=1., nugget=0.5)
    model.fit(locs1, vals1, iters=200)

    # # MCMC.
    # model = Model(
    #     gp,
    #     parameters = dict(b1=0., b2=0., b3=0., range=1., nugget=0.5),
    #     verbose = True).mcmc(locs1, vals1, burnin=500, samples=500)

    # Interpolate using GP.
    N = 20
    xx, yy = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy], axis=-1).reshape([-1, 2])

    mean, var = model.predict(locs2)
    mean2, var2 = model.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)

def test_explicit_trend():
    # Create 500 random locations in a square centered on the origin.
    locs1 = np.random.uniform(-1., 1., [500, 2])

    p = Parameters(beta=[1., 0.5, 0.5, 0.25], range=0.5, nugget=1.)

    # Initialize featurizer of location for trends.
    def trend_terms(x, y): return x, y, x*y
    featurizer = NormalizingFeaturizer(trend_terms, locs1)
    gp = GP(Trend(featurizer, beta=p.beta),
            krn.SquaredExponential(range=p.range, sill=1.) + krn.Noise(nugget=p.nugget))

    # Generating GP.
    model = Model(gp)

    # Generate data.
    vals1 = model.generate(locs1).vals

    # Fit GP.
    model.set(beta=[0., 0., 0., 0.], range=1., nugget=0.5)
    model.fit(locs1, vals1, iters=200)

    # # MCMC.
    # model = Model(
    #     gp,
    #     parameters = dict(beta=[0., 0., 0., 0.], range=1., nugget=0.5),
    #     verbose = True).mcmc(locs1, vals1, burnin=500, samples=500)

    # Interpolate using GP.
    N = 20
    xx, yy = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy], axis=-1).reshape([-1, 2])

    mean, var = model.predict(locs2)
    mean2, var2 = model.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)



from types import SimpleNamespace
from dataclasses import dataclass, replace
from typing import Dict
import tensorflow as tf
import numpy as np
from scipy.special import logit


__all__ = ['Parameter', 'Parameters']

def Parameters(**kw):
    return SimpleNamespace(**{k:Parameter(k, v) for k, v in kw.items()})

@dataclass
class Parameter:
    name: str
    value: float
    lo: float = np.nan
    hi: float = np.nan
    underlying: tf.Variable = None

    def update_bounds(self, lo: float, hi: float):
        if np.isnan(self.lo):
            self.lo = lo
        else:
            assert self.lo == lo, f'Conflicting bounds for parameter {self.name}'
        if np.isnan(self.hi):
            self.hi = hi
        else:
            assert self.hi == hi, f'Conflicting bounds for parameter {self.name}'

    def bounding(self):
        return \
            ('u' if self.lo == float('-inf') else 'b') + \
            ('u' if self.hi == float('inf') else 'b')

    def create_tf_variable(self):
        """Create TF variable for underlying parameter or update it"""
        # Create underlying parameter.
        b = self.bounding()
        if b == 'bb':
            init = logit((self.value - self.lo) / (self.hi - self.lo))
        elif b == 'bu':
            init = np.log(self.value - self.lo)
        elif b == 'ub':
            init = -np.log(self.hi - self.value)
        else:
            init = self.value

        if self.underlying is None:
            self.underlying = tf.Variable(init, name=self.name, dtype=tf.float32)
        else:
            self.underlying.assign(init)

    def surface(self):
        """ Create tensor for surface parameter"""
        # Create surface parameter.
        b = self.bounding()
        v = self.underlying
        if b == 'bb':
            v = tf.math.sigmoid(v) * (self.hi - self.lo) + self.lo
        elif b == 'bu':
            v = tf.exp(v) + self.lo
        elif b == 'ub':
            v = self.hi - tf.exp(-v)
        else:
            v = v + tf.constant(0.)
        return v

    def update_value(self):
        self.value = self.surface().numpy()

# TODO: cache surface somehow
# TODO: replace with map nested?
def get_parameter_values(blob: object):
    """
    For each Parameter encountered in the nested blob,
    replace it with its surface tensor.
    """
    if isinstance(blob, dict):
        return {k: get_parameter_values(a) for k, a in blob.items()}
    elif isinstance(blob, (list, tuple)):
        return [get_parameter_values(a) for a in blob]
    elif isinstance(blob, Parameter):
        return blob.surface()
    elif isinstance(blob, str):
        raise ValueError(f'Bad parameter {blob} is a string')
    else:
        return blob

def ppp(param):
    """Positive paper parameter (maybe)."""
    if isinstance(param, Parameter):
        param.update_bounds(0., np.inf)
        return {param.name: param}
    else:
        return {}

def upp(param):
    """Unbounded paper parameter (maybe)."""
    if isinstance(param, Parameter):
        param.update_bounds(-np.inf, np.inf)
        return {param.name: param}
    else:
        return {}

def bpp(param, lo, hi):
    """Bounded paper parameter (maybe)."""
    if isinstance(param, Parameter):
        param.update_bounds(lo, hi)
        return {param.name: param}
    else:
        return {}

def ppp_list(beta):
    if isinstance(beta, (list, tuple)):
        return {k: p for s in beta for k, p in ppp(s).items()}
    elif isinstance(beta, Parameter):
        return ppp(beta)
    else:
        return {}



import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.distance import cdist, pdist
from scipy.stats import binned_statistic
import scipy.stats as stats
from scipy.optimize import curve_fit
import pandas as pd

__all__ = ['Krige']

########################################################################
# Cutoff distance function using Pythagorean Theorem.
# Potentially need to find a better citation for this and a better automated method.
# Myers, D. E. (1991). On variogram estimation. The frontiers of 
# statistical scientific theory & industrial applications, 2, 261-266.

def cutoff_dist_func(x):
    
    '''
        Parameters:
                x : n-dim array
                    Locations of input data.
                    
        Returns:
                cutoff : float
                    The maximum lag distance to use in fitting the variogram.
                    Found using Pythagorean Theorem to roughly find one half the distance across the study area.
        
    '''
    
    a2 = np.square(x[:, 0].max() - x[:, 0].min())
    b2 = np.square(x[:, 1].max() - x[:, 1].min())

    cutoff = np.sqrt(a2 + b2) / 2

    return cutoff
    

#####################################################################   
# Variogram models.

def linear(D, *parameter_vals):
    slope, nugget = parameter_vals
    return slope * D + nugget

def gaussian(D, *parameter_vals):
    vrange, sill, nugget = parameter_vals
    return sill * (1. - np.exp(-np.square(D / (vrange)))) + nugget

def spherical(D, *parameter_vals):
    vrange, sill, nugget = parameter_vals
    return np.piecewise(D, [D <= vrange, D > vrange],
                    [lambda x: sill * ((3. * x) / (2. * vrange) - (x**3.) 
                                       / (2. * vrange**3.)) + nugget, sill + nugget])




class Krige():
    
    def __init__(self, 
                 x1, u1, 
                 bins,
                 variogram_func=None, 
                 variogram_params=None, 
                 cutoff_dist='auto',
                 featurization=None,
                 projection=None,
                 show_plots=True, 
                 verbose=True,
                 ):
        
        '''
        Parameters:
                x1 : n-dim array
                    Locations of input data.
                    
                u1 : 1-d array
                    Values to be kriged.
                
                bins : int or None
                    The number of bins to use on the variogram cloud.
                    If None, variogram function is fit to the variogram cloud
                    and is not binned first.
                    
                variogram_func : str
                     Name of the variogram model to use in the kriging. 
                     Should be 'linear', 'gaussian', or 'spherical'.
                     Default is 'gaussian'.

                cutoff_dist : str or int, optional
                    The maximum lag distance to include in variogram modeling.
                    
                featurization : function, optional
                    Should be a function that takes x1 (n-dim array of input data) 
                    and returns the coordinates, i.e., x, y, x**2, y**2.
                    Example below.
                    Default is None.
                    
                project : function, opt
                    A function that takes multiple vectors, and returns
                    a tuple of projected vectors.

                epsg_proj : str
                    The projected coordinate system to use. Ignored if project=False.
                    Default is 'EPSG:3310' (California Albers).

                show_plots : boolean, optional
                    Whether or not to show variogram plots.
                    Default is True.

                verbose : boolean, optional
                    Whether or not to print parameters.
                    Default is True.


        Performs experimental variogram calculation, bins data, and fits variogram model to estimate variogram parameters.       
                 
        Performs ordinary and universal kriging in up to 3 spatial dimensions.
        
        
        Trend model example:
        def featurization(x1):    
            return x1[:, 0], x1[:, 1], x1[:, 0]**2, x1[:, 1]**2
        
        
        '''

        super().__init__(projection=projection)
        
        # Save original x1 for stats function.
        self.x1_original = x1
        
        # Check and change the shape of u1 if needed (for pdist).
        if u1.ndim == 1:
            self.u1 = u1
        elif u1.shape[1] == 1:
            self.u1 = u1.reshape(-1)
        else:
            raise ValueError("Check dimensions of 'u1'.")
            
        # Projection.
        self.x1 = self.project(x1)

        # Variogram.
        if variogram_func == 'gaussian':
            self.variogram_func = gaussian
        elif variogram_func == 'spherical':
            self.variogram_func = spherical
        elif variogram_func == 'linear':
            self.variogram_func = linear
        else:  
            raise ValueError("Variogram function must be 'linear', 'gaussian', or 'spherical'.")


        if cutoff_dist == 'auto':
            self.cutoff_dist = cutoff_dist_func(self.x1)
        else:
            self.cutoff_dist = cutoff_dist


        self.verbose = verbose
        self.show_plots = show_plots
        self.bins = bins
        self.featurization = featurization

            
        # Lags and semivariance.
        dist = pdist(self.x1, metric='euclidean')
        gamma = 0.5 * pdist(self.u1.reshape(-1, 1), metric='sqeuclidean')


        if self.bins != None:  # Use bins.
            # Variogram cloud calculation.
            bin_means, bin_edges, binnumber = binned_statistic(dist, gamma, 
                                                               statistic='mean', 
                                                               bins=self.bins, 
                                                               range=[dist.min(), 
                                                                      self.cutoff_dist])
            bin_width = (bin_edges[1] - bin_edges[0])
            bin_centers = bin_edges[1:] - bin_width / 2

            # Bin counts calculation.
            bin_count, bin_count_edges, bin_count_number = binned_statistic(dist, gamma, 
                                                                            statistic='count', 
                                                                            bins=self.bins)
            bin_count_width = (bin_count_edges[1] - bin_count_edges[0])
            bin_count_centers = bin_count_edges[1:] - bin_count_width/2

        if self.show_plots == True:
            if self.bins == None:

                # Variogram cloud plot.
                plt.figure(dpi=100)
                plt.scatter(dist, gamma, ec='C0', fc='none', alpha=0.3)
                plt.ylabel('$\gamma(h)$')
                plt.xlabel('$h$')
                plt.grid(alpha=0.4)
                plt.show()

            if self.bins != None:
                
                # Variogram cloud plot.
                plt.figure(dpi=100)
                plt.scatter(dist, gamma, ec='C0', fc='none', alpha=0.3)
                plt.hlines(bin_means, bin_edges[:-1], bin_edges[1:], zorder=1, color='k')
                plt.scatter(bin_centers, bin_means, ec='k', lw=0.5)
                plt.ylabel('$\gamma(h)$')
                plt.xlabel('$h$')
                plt.grid(alpha=0.4)
                plt.show()

                # Bin counts plot.
                plt.figure(dpi=100)
                plt.hlines(bin_count, bin_count_edges[:-1], bin_count_edges[1:], zorder=1, color='k')
                plt.scatter(bin_count_centers, bin_count, ec='k', lw=0.5)
                plt.ylabel('Bin count')
                plt.xlabel('$h$')
                plt.grid(alpha=0.4)
                plt.show()

        
        
        ############ Fit the variogram model.
        
        if not variogram_params:  # Fit to the data.
            
            if self.bins == None:  # Fit the variogram cloud.
                
                if self.variogram_func == gaussian or self.variogram_func == spherical:

                    # Initial guess for the parameters.
                    p0 = [0.25 * np.max(dist),            # range
                          np.max(gamma) - np.min(gamma),  # sill
                          np.min(gamma) + 1e-6]           # nugget

                    # Bounds with constraints.
                    bounds = [(1e-6, 1e-6, 1e-6), (np.inf, np.inf, np.inf)]

                    # Apply the cutoff.
                    dist_cut = np.where(dist < self.cutoff_dist, dist, np.nan)
                    gamma_cut = np.where(dist < self.cutoff_dist, gamma, np.nan)
                    
                    # Remove nans for curve_fit.
                    dist_cut = dist_cut[~np.isnan(dist_cut)]
                    gamma_cut = gamma_cut[~np.isnan(gamma_cut)]
                    
                    popt, pcov = curve_fit(self.variogram_func, dist_cut, gamma_cut, p0=p0, bounds=bounds)

                    if self.show_plots == True:
                        
                        # Calculate 2d kde to help confirm cutoff.
                        xi = np.linspace(np.min(dist), np.max(dist), 60)
                        yi = np.linspace(np.min(gamma), np.max(gamma), 60)
                        xi, yi = np.meshgrid(xi, yi)
                        xyi = np.stack([xi.reshape(-1), yi.reshape(-1)], axis=1).T

                        kde = stats.gaussian_kde([dist, gamma])
                        z = kde.evaluate(xyi)
                        z = z.reshape(len(xi), len(yi))

                        plt.figure(dpi=100)
                        plt.pcolormesh(xi, yi, z, cmap=plt.cm.Blues, shading='auto')
                        plt.ylabel('$\gamma(h)$')
                        plt.xlabel('$h$')
                        plt.grid(alpha=0.4)
                        plt.show()
                        
                        # With cutoff and variogram model.
                        xnew = np.linspace(np.min(dist_cut), self.cutoff_dist, 100)
                        plt.figure(dpi=100)
                        plt.scatter(dist_cut, gamma_cut, fc='none', ec='C1', lw=0.5, alpha=0.3)
                        plt.plot(xnew, self.variogram_func(xnew, *popt), color='k')
                        plt.ylabel('$\gamma(h)$')
                        plt.xlabel('$h$')
                        plt.grid(alpha=0.4)
                        plt.show()
                        

                    vrange = popt[0]
                    sill = popt[1]
                    nugget = popt[2]


                    if self.verbose == True:
                        print('variogram model: {}'.format(variogram_func))
                        print('cutoff: {:.2f}'.format(self.cutoff_dist))
                        print('range: {:.5f}'.format(vrange))
                        print('sill: {:.5f}'.format(sill))
                        print('nugget: {:.5f}'.format(nugget))
                        print('full sill: {:.5f}'.format(sill + nugget))

                    self.parameter_vals = [vrange, sill, nugget]

                elif self.variogram_func == linear:

                    # Initial guess for the parameters.
                    p0 = [(np.max(dist) - np.min(dist)) / (np.max(gamma) - np.min(gamma)),  # slope
                          np.min(gamma) + 1e-6]                                             # nugget

                    # Bounds with constraints.
                    bounds = [(1e-6, 1e-6), (np.inf, np.inf)]

                    # Apply the cutoff.
                    dist_cut = np.where(dist < self.cutoff_dist, dist, np.nan)
                    gamma_cut = np.where(dist < self.cutoff_dist, gamma, np.nan)
                    
                    # Remove nans for curve_fit.
                    dist_cut = dist_cut[~np.isnan(dist_cut)]
                    gamma_cut = gamma_cut[~np.isnan(gamma_cut)]
                    
                    popt, pcov = curve_fit(self.variogram_func, dist_cut, gamma_cut, p0=p0, bounds=bounds)

                    if self.show_plots == True:

                        # Calculate 2d kde to help confirm cutoff.
                        xi = np.linspace(np.min(dist), np.max(dist), 60)
                        yi = np.linspace(np.min(gamma), np.max(gamma), 60)
                        xi, yi = np.meshgrid(xi, yi)
                        xyi = np.stack([xi.reshape(-1), yi.reshape(-1)], axis=1).T

                        kde = stats.gaussian_kde([dist, gamma])
                        z = kde.evaluate(xyi)
                        z = z.reshape(len(xi), len(yi))

                        plt.figure(dpi=100)
                        plt.pcolormesh(xi, yi, z, cmap=plt.cm.Blues, shading='auto')
                        plt.ylabel('$\gamma(h)$')
                        plt.xlabel('$h$')
                        plt.grid(alpha=0.4)
                        plt.show()
                        
                        # With cutoff and variogram model.
                        xnew = np.linspace(np.min(dist_cut), self.cutoff_dist, 100)
                        plt.figure(dpi=100)
                        plt.scatter(dist_cut, gamma_cut, fc='none', ec='C1', lw=0.5, alpha=0.3)
                        plt.plot(xnew, self.variogram_func(xnew, *popt), color='k')
                        plt.ylabel('$\gamma(h)$')
                        plt.xlabel('$h$')
                        plt.grid(alpha=0.4)
                        plt.show()
                        

                    slope = popt[0]
                    nugget = popt[1]

                    if self.verbose == True:
                        print('variogram model: {}'.format(variogram_func))
                        print('cutoff: {:.2f}'.format(self.cutoff_dist))
                        print('slope: {:.5f}'.format(slope))
                        print('nugget: {:.5f}'.format(nugget))


                    self.parameter_vals = [slope, nugget]

            
            
            
            elif self.bins != None:  # Fit the binned data.
                if self.variogram_func == gaussian or self.variogram_func == spherical:

                    # Initial guess for the parameters.
                    p0 = [0.25 * np.max(bin_centers),                 # range
                              np.max(bin_means) - np.min(bin_means),  # sill
                              np.min(bin_means)]                      # nugget

                    # Bounds with constraints.
                    bounds = [(1e-6, 1e-6, 1e-6), (np.inf, np.inf, np.inf)]

                    popt, pcov = curve_fit(self.variogram_func, bin_centers, bin_means, p0=p0, bounds=bounds)

                    if self.show_plots == True:

                        # Fit variogram plot.
                        plt.figure(dpi=100)
                        plt.scatter(bin_centers, bin_means, c='C1', ec='k', lw=0.5)
                        plt.plot(bin_centers, self.variogram_func(bin_centers, *popt), color='k')
                        plt.ylabel('$\gamma(h)$')
                        plt.xlabel('$h$')
                        plt.grid(alpha=0.4)
                        plt.show()

                    vrange = popt[0]
                    sill = popt[1]
                    nugget = popt[2]


                    if self.verbose == True:
                        print('variogram model: {}'.format(variogram_func))
                        print('cutoff: {:.2f}'.format(self.cutoff_dist))
                        print('range: {:.5f}'.format(vrange))
                        print('sill: {:.5f}'.format(sill))
                        print('nugget: {:.5f}'.format(nugget))
                        print('full sill: {:.5f}'.format(sill + nugget))

                    self.parameter_vals = [vrange, sill, nugget]

                elif self.variogram_func == linear:

                    # Initial guess for the parameters.
                    p0 = [(np.max(bin_centers) - np.min(bin_centers)) / (np.max(bin_means) - np.min(bin_means)),  # slope
                          np.min(bin_means)]                                                                      # nugget

                    # Bounds with constraints.
                    bounds = [(1e-6, 1e-6), (np.inf, np.inf)]

                    popt, pcov = curve_fit(self.variogram_func, bin_centers, bin_means, p0=p0, bounds=bounds)

                    if self.show_plots == True:

                        # Fit variogram plot.
                        plt.figure(dpi=100)
                        plt.scatter(bin_centers, bin_means, c='C1', ec='k', lw=0.5)
                        plt.plot(bin_centers, self.variogram_func(bin_centers, *popt), color='k')
                        plt.ylabel('$\gamma(h)$')
                        plt.xlabel('$h$')
                        plt.grid(alpha=0.4)
                        plt.show()

                    slope = popt[0]
                    nugget = popt[1]

                    if self.verbose == True:
                        print('variogram model: {}'.format(variogram_func))
                        print('cutoff: {:.2f}'.format(self.cutoff_dist))
                        print('slope: {:.5f}'.format(slope))
                        print('nugget: {:.5f}'.format(nugget))


                    self.parameter_vals = [slope, nugget]


        else:  # Use the given variogram parameters.
            self.parameter_vals = variogram_params
            
            # Apply the cutoff.
            dist_cut = np.where(dist < self.cutoff_dist, dist, np.nan)
            gamma_cut = np.where(dist < self.cutoff_dist, gamma, np.nan)

            # Remove nans for curve_fit.
            dist_cut = dist_cut[~np.isnan(dist_cut)]
            gamma_cut = gamma_cut[~np.isnan(gamma_cut)]
            
            
            if self.show_plots == True:

                if self.bins == None:
                    
                    # Calculate 2d kde to help confirm cutoff.
                    xi = np.linspace(np.min(dist), np.max(dist), 60)
                    yi = np.linspace(np.min(gamma), np.max(gamma), 60)
                    xi, yi = np.meshgrid(xi, yi)
                    xyi = np.stack([xi.reshape(-1), yi.reshape(-1)], axis=1).T
                    
                    kde = stats.gaussian_kde([dist, gamma])
                    z = kde.evaluate(xyi)
                    z = z.reshape(len(xi), len(yi))

                    plt.figure(dpi=100)
                    plt.pcolormesh(xi, yi, z, cmap=plt.cm.Blues, shading='auto')
                    plt.ylabel('$\gamma(h)$')
                    plt.xlabel('$h$')
                    plt.grid(alpha=0.4)
                    plt.show()

                    # With cutoff and variogram model.
                    xnew = np.linspace(np.min(dist_cut), self.cutoff_dist, 100)
                    plt.figure(dpi=100)
                    plt.scatter(dist_cut, gamma_cut, fc='none', ec='C1', lw=0.5, alpha=0.3)
                    plt.plot(xnew, self.variogram_func(xnew, *self.parameter_vals), color='k')
                    plt.ylabel('$\gamma(h)$')
                    plt.xlabel('$h$')
                    plt.grid(alpha=0.4)
                    plt.show()
                    
                else:
                    
                    # Fit variogram plot.
                    plt.figure(dpi=100)
                    plt.scatter(bin_centers, bin_means, fc='C1', ec='k', lw=0.5)
                    plt.plot(bin_centers, self.variogram_func(bin_centers, *self.parameter_vals), color='k')
                    plt.ylabel('$\gamma(h)$')
                    plt.xlabel('$h$')
                    plt.grid(alpha=0.4)
                    plt.show()

            if self.variogram_func == gaussian or self.variogram_func == spherical:
                vrange = self.parameter_vals[0]
                sill = self.parameter_vals[1]
                nugget = self.parameter_vals[2]

                if self.verbose == True:
                    print('variogram model: {}'.format(variogram_func))
                    print('cutoff: {:.2f}'.format(self.cutoff_dist))
                    print('range: {:.5f}'.format(vrange))
                    print('sill: {:.5f}'.format(sill))
                    print('nugget: {:.5f}'.format(nugget))
                    print('full sill: {:.5f}'.format(sill + nugget))

            elif self.variogram_func == linear:
                slope = self.parameter_vals[0]
                nugget = self.parameter_vals[1]

                if self.verbose == True:
                    print('variogram model: {}'.format(variogram_func))
                    print('cutoff: {:.2f}'.format(self.cutoff_dist))
                    print('slope: {:.5f}'.format(slope))
                    print('nugget: {:.5f}'.format(nugget))
 
            
        
    
####################################################

    def predict(self, x2_pred):
        
        '''
        Parameters:
                x2 : n-dim array
                    Locations to make kriging predictions.
        
        Returns:
                u2_mean : float
                    Kriging mean.
                    
                u2_var : float
                    Kriging variance.
        
        
        Performs ordinary or universal kriging using the estimated variogram parameters.
        
        '''
        
        self.x2 = self.project(x2_pred)
        
        n1 = len(self.x1)
        n2 = len(self.x2)
        
        # Universal krige.
        if self.featurization:            
            # Ax = b with a trend.
            
            # Build A
            D1 = cdist(self.x1, self.x1)

            drift_data = np.array(list(self.featurization(self.x1)))

            An = n1 + 1 + drift_data.shape[0]

            A = np.zeros((An, An))
            A[:n1, :n1] = -self.variogram_func(D1, *self.parameter_vals)
            np.fill_diagonal(A, 0.)
            A[n1, :n1] = 1.
            A[:n1, n1] = 1.
            A[n1, n1] = 0.

            # Add in the trend for A.
            for i in range(drift_data.shape[0]):
                A[n1+i+1, :n1] = drift_data[i]
                A[:n1, n1+i+1] = drift_data[i]

            # Build b.
            D2 = cdist(self.x2, self.x1)
            b = np.zeros((D2.shape[0], D2.shape[1] + 1 + drift_data.shape[0]))
            b[:n2, :n1] = -self.variogram_func(D2, *self.parameter_vals)
            b = b.T
            b[n1, :] = 1.

            # Add the trend for b.
            drift_pred = np.array(list(self.featurization(self.x2)))

            for i in range(drift_pred.shape[0]):
                b[n1+1+i, :] = drift_pred[i]

            # Solve.
            x = np.linalg.solve(A, b)

            u2_mean = np.tensordot(self.u1, x[:n1], axes=1)
            u2_var = np.sum(x.T * -b.T, axis=1)

            return u2_mean, u2_var

        
        # Ordinary krige.
        else:
            # Ax = b.

            # Build A.
            D1 = cdist(self.x1, self.x1)
            A = np.zeros((n1+1, n1+1))
            A[:n1, :n1] = -self.variogram_func(D1, *self.parameter_vals)
            np.fill_diagonal(A, 0.)
            A[n1, :] = 1.
            A[:, n1] = 1.
            A[n1, n1] = 0.

            # Build b.
            D2 = cdist(self.x2, self.x1)
            b = np.zeros((D2.shape[0], D2.shape[1]+1))
            b[:n2, :n1] = -self.variogram_func(D2, *self.parameter_vals)
            b = b.T
            b[n1, :] = 1.

            # Solve.
            x = np.linalg.solve(A, b)

            u2_mean = np.tensordot(self.u1, x[:n1], axes=1)
            u2_var = np.sum(x.T * -b.T, axis=1)

            return u2_mean, u2_var
        
    #########################################################
    # Access the projected coords of the input data.
    def get_projected(self):
        return self.x1[:, 0], self.x1[:, 1]
    
    #########################################################   
    # Provide some stats on the model.
    def stats(self):         

        u2_mean_for_u1, u2_var_for_u1 = self.predict(self.x1_original)
        
        self.residuals = self.u1 - u2_mean_for_u1     
        
        self.res_mean = np.mean(self.residuals)
        self.res_std = np.std(self.residuals)
        self.res_skew = stats.skew(self.residuals)
        self.res_kurt = stats.kurtosis(self.residuals)
            
        if self.show_plots == True:

            plt.figure()
            plt.hist(self.u1, label='Input values')
            plt.ylabel('Count')
            plt.xlabel('Input values')
            plt.show()

            plt.figure()
            plt.hist(self.residuals, label='Residuals')
            plt.ylabel('Count')
            plt.xlabel('Residuals')
            plt.show()

        if self.verbose == True:
            print('Residual mean: {:.3e}'.format(self.res_mean))
            print('Residual standard deviation: {:.3f}'.format(self.res_std))
            print('Residual skewness: {:.3f}'.format(self.res_skew))
            print('Residual kurtosis: {:.3f}'.format(self.res_kurt))

    def get_residuals(self):
        return self.residuals

    def get_residual_moments(self):
        return dict(mean=self.res_mean, 
                    std=self.res_std, 
                    skew=self.res_skew, 
                    kurt=self.res_kurt)



import time
from collections import defaultdict
from dataclasses import dataclass, replace
from typing import Callable, Dict, List, Optional, Union
import numpy as np
from scipy.special import expit, logit
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'

# Tensorflow is extraordinarily noisy. Catch warnings during import.
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import tensorflow as tf
    import tensorflow_probability as tfp
    tfd = tfp.distributions

from tensorflow.core.function.trace_type import default_types
import tensorflow_probability as tfp

from . import mean as mn
from . import kernel as krn
from .metric import Euclidean, PerAxisDist2
from .op import Op
from .op import SingletonTraceType
from .param import get_parameter_values, ppp, upp, bpp
from .param import Parameter

MVN = tfp.distributions.MultivariateNormalTriL

__all__ = ['featurizer', 'GP', 'Mix', 'Model', 'Featurizer', 'NormalizingFeaturizer', 'StratigraphicWarp']

@dataclass
class GP:
    mean: mn.Trend = None
    kernel: krn.Kernel = None

    def __post_init__(self):
        if self.mean is None or self.mean == 0:
            self.mean = mn.ZeroTrend()
        assert self.kernel is not None

    def __add__(self, other):
        return GP(self.mean + other.mean, self.kernel + other.kernel)

    def __tf_tracing_type__(self, context):
        return SingletonTraceType(self)

    def gather_vars(self):
        return self.mean.gather_vars() | self.kernel.gather_vars()

def Mix(inputs, weights=None):
    return GP(
        mn.Mix([i.mean for i in inputs], weights), 
        krn.Mix([i.kernel for i in inputs], weights))

class Warp:
    def __call__(self, locs, prep):
        """
        `locs` is numpy.

        Returns a WarpLocations.
        """
        pass

    def gather_vars(self):
        pass

class WarpLocations(Op):
    def __init__(self, warped_locs):
        self.warped_locs = warped_locs
        super().__init__({}, {})

    def __call__(self, e):
        return tf.cast(self.warped_locs, dtype=tf.float32)

    def __tf_tracing_type__(self, context):
        return SingletonTraceType(self)

    def gather_vars(self):
        return {}

class NoWarp:
    def __call__(self, locs):
        return WarpLocations(locs)

    def gather_vars(self):
        return {}

class CrazyWarp(Warp):
    def __call__(self, locs):
        if locs.shape[-1] == 4:
            locs += e(np.sin(locs[:, 1])) * np.array([0., 0., 1., 0.])
        return WarpLocations(locs)

    def gather_vars(self):
        return {}

# @tf.function()
def interpolate_1d_tf(src, tgt, x):
    """
    `src`: (batch, breaks)
    `tgt`: (batch, breaks)
    `x`  : (batch)
    """
    x_shape = tf.shape(x)
    x = tf.reshape(x, [-1, 1]) # (batch, 1)
    bucket = tf.searchsorted(src, x)
    bucket = tf.clip_by_value(bucket - 1, 0, tf.shape(tgt)[0] - 2)
    # print('====================== x bucket')
    # print(tf.shape(x))
    # print(tf.shape(bucket))
    src0 = tf.gather(src, bucket, batch_dims=1)
    src1 = tf.gather(src, bucket + 1, batch_dims=1)
    tgt0 = tf.gather(tgt, bucket, batch_dims=1)
    tgt1 = tf.gather(tgt, bucket + 1, batch_dims=1)
    # print('====================== src0 sr1 tgt0 tgt1')
    # print(tf.shape(src0))
    # print(tf.shape(src1))
    # print(tf.shape(tgt0))
    # print(tf.shape(tgt1))
    xout = ((x - src0) * tgt1 + (src1 - x) * tgt0) / (src1 - src0)
    return tf.reshape(xout, x_shape)

@tf.function
def relax(s, t, distort):
    xi = distort / (1 - distort)
    ds = s[:, 1:] - s[:, :-1]
    x = s

    # print('-------------------- t')
    # print(t)
    # print('-------------------- x')
    # print(x[:6, :])
    for i in range(5):
        # Compute objective.
        dx = x[:, 1:] - x[:, :-1]
        dxds = dx / ds
        a = tf.math.log(dxds)
        # obj = tf.reduce_sum(tf.math.square(a), axis=-1) \
        #     + tf.reduce_sum(xi * tf.math.square(x - t), axis=-1)

        # print('-------------------- obj')
        # print(obj[:6])

        # Compute gradient.
        # print('-------------------- g')
        # print(g[:6, :])
        g = a / dx
        zg = tf.pad(g, [[0, 0], [1, 0]])
        gz = tf.pad(g, [[0, 0], [0, 1]])
        grad = 2 * (zg - gz + xi * (x - t))
        # print('-------------------- -grad')
        # print(-grad)

        # Compute hessian as tridiagonal matrix.
        h = (1 - tf.math.log(dxds)) / tf.square(dxds) / tf.square(ds)
        zh = tf.pad(h, [[0, 0], [1, 0]])
        hz = tf.pad(h, [[0, 0], [0, 1]])
        diag = 2 * (hz + zh + xi)
        offdiag = -2 * h
        zo = tf.pad(offdiag, [[0, 0], [1, 0]])
        oz = tf.pad(offdiag, [[0, 0], [0, 1]])
        compact = tf.stack([oz, diag, zo], axis=-2)
        
        # Newton's method.
        x -= tf.linalg.tridiagonal_solve(compact, grad)

        # print('-------------------- x')
        # print(x[:6, :])

    return x

class TweeningStratigraphicWarp(Warp):
    def __init__(self, surface_functions, surface_targets, tween):
        self.surface_functions = surface_functions
        self.surface_targets = surface_targets
        self.tween = tween

    def __call__(self, locs):
        sources = [f(locs[..., :2]) for f in self.surface_functions]
        sources = tf.sort(tf.cast(tf.stack(sources, axis=-1), tf.float32))
        targets = tf.constant(self.surface_targets, dtype=tf.float32)

        locs = tf.cast(locs, tf.float32)
        zout = interpolate_1d_tf(sources, targets, locs[..., 2])
        out = tf.stack([locs[..., 0], locs[..., 1], zout, locs[..., 3]], axis=-1)
        return TweeningStratigraphicWarpLocations(locs, out.numpy(), self.tween)

    def gather_vars(self):
        return bpp(self.tween, 0., 1.)

class TweeningStratigraphicWarpLocations(Op):
    def __init__(self, locs, warped_locs, tween):
        self.locs = locs
        self.warped_locs = warped_locs

        fa = dict(tween=tween)
        super().__init__(fa, {})

    def __call__(self, e):
        tween = e['tween']
        locs0 = tf.cast(self.locs, dtype=tf.float32)
        locs1 = tf.cast(self.warped_locs, dtype=tf.float32)
        return locs0 * (1. - tween) + locs1 * tween

    def __tf_tracing_type__(self, context):
        return SingletonTraceType(self)

    def gather_vars(self):
        return bpp(self.fa['tween'], 0., 1.)

class StratigraphicWarp(Warp):
    def __init__(self, surface_functions, surface_targets, distort):
        self.surface_functions = surface_functions
        self.surface_targets = surface_targets
        self.distort = distort

    def __call__(self, locs):
        sources = [f(locs[..., :2]) for f in self.surface_functions]
        sources = tf.sort(tf.cast(tf.stack(sources, axis=-1), tf.float32))
        targets = tf.constant(self.surface_targets, dtype=tf.float32)
        locs = tf.cast(locs, tf.float32)
        return StratigraphicWarpLocations(locs, sources, targets, self.distort)

    def gather_vars(self):
        return bpp(self.distort, 0., 1.)

class StratigraphicWarpLocations(Op):
    def __init__(self, locs, sources, targets, distort):
        self.locs = locs
        self.sources = sources
        self.targets = targets

        fa = dict(distort=distort)
        super().__init__(fa, {})

    def __call__(self, e):
        distort = e['distort']

        locs0 = self.locs
        s = self.sources
        t = self.targets
        r = relax(s, t, distort)
        z = interpolate_1d_tf(s, r, locs0[:, 2])
        locs1 = tf.stack([locs0[:, 0], locs0[:, 1], z, locs0[:, 3]], axis=-1)

        return locs1

    def __tf_tracing_type__(self, context):
        return SingletonTraceType(self)

    def gather_vars(self):
        return bpp(self.fa['tween'], 0., 1.)

class NormalizingFeaturizer:
    """
    Produces featurized locations (F matrix) and remembers normalization
    parameters.  Adds an intercept feature.
    """
    def __init__(self, featurization, locs):
        self.unnorm_featurizer = Featurizer(featurization)
        F_unnorm = self.unnorm_featurizer(locs)
        self.unnorm_mean = tf.reduce_mean(F_unnorm, axis=0)
        self.unnorm_std = tf.math.reduce_std(F_unnorm, axis=0)

    def __call__(self, locs):
        ones = tf.ones([tf.shape(locs)[0], 1], dtype=tf.float32)
        F_unnorm = self.unnorm_featurizer(locs)
        F_norm = (F_unnorm - self.unnorm_mean) / self.unnorm_std
        return tf.concat([ones, F_norm], axis=1)

class Featurizer:
    """
    Produces featurized locations (F matrix).
    """
    def __init__(self, featurization):
        self.featurization = featurization

    def __call__(self, locs):
        locs = tf.cast(locs, tf.float32)
        if self.featurization is None: # No features.
            return tf.ones([tf.shape(locs)[0], 0], dtype=tf.float32)

        feats = self.featurization(*tf.unstack(locs, axis=1))
        if isinstance(feats, tuple): # One or many features.
            if len(feats) == 0:
                return tf.ones([tf.shape(locs)[0], 0], dtype=tf.float32)
            else:
                feats = self.featurization(*tf.unstack(locs, axis=1))
                feats = [tf.broadcast_to(tf.cast(f, tf.float32), [tf.shape(locs)[0]]) for f in feats]
                return tf.stack(feats, axis=1)
        else: # One feature.
            return e(feats)

def featurizer(normalize=None):
    def helper(f):
        if normalize is None:
            return Featurizer(f)
        else:
            return NormalizingFeaturizer(f, normalize)
    return helper

def e(x, a=-1):
    return tf.expand_dims(x, a)

# @tf.function
def gp_covariance(gp, locs, cats):
    return gp_covariance2(gp, locs, cats, locs, cats, 0)

# @tf.function
def gp_covariance2(gp, locs1, cats1, locs2, cats2, offset):
    """
    `offset` is i2-i1, where i1 and i2 are the starting indices of locs1
    and locs2.  It is used to create the diagonal non-zero elements
    of a Noise covariance function.  An non-zero offset results in a
    covariance matrix with non-zero entries along an off-center diagonal.
    """

    # assert np.all(cats1 == np.sort(cats1)), '`cats1` must be in non-descending order'
    # assert np.all(cats2 == np.sort(cats2)), '`cats2` must be in non-descending order'

    cache = {}
    cache['offset'] = offset
    cache['locs1'] = locs1
    cache['locs2'] = locs2
    cache['cats1'] = cats1
    cache['cats2'] = cats2
    cache['per_axis_dist2'] = PerAxisDist2().run(cache)
    cache['euclidean'] = Euclidean().run(cache)

    M = gp.mean.run(cache)
    C = gp.kernel.run(cache)
    M = tf.cast(M, tf.float64)
    C = tf.cast(C, tf.float64)
    return M, C

def mvn_log_pdf(u, m, cov):
    """Log PDF of a multivariate gaussian."""
    u_adj = u - m
    logdet = tf.linalg.logdet(2 * np.pi * cov)
    quad = tf.matmul(e(u_adj, 0), tf.linalg.solve(cov, e(u_adj, -1)))[0, 0]
    return tf.cast(-0.5 * (logdet + quad), tf.float32)

# @tf.function
def gp_log_likelihood(data, gp):
    m, S = gp_covariance(gp, data['warplocs'].run({}), data['cats'])
    u = tf.cast(data['vals'], tf.float64)
    return mvn_log_pdf(u, m, S)

def gp_train_step(
    optimizer,
    data,
    parameters: Dict[str, Parameter],
    gp,
    reg=None
):
    with tf.GradientTape() as tape:
        ll = gp_log_likelihood(data, gp)

        if reg:
            # TODO: Put in cache later.
            reg_penalty = reg.run({})
        else:
            reg_penalty = 0.

        loss = -ll + reg_penalty

    up = [p.underlying for p in parameters.values()]
    gradients = tape.gradient(loss, up)
    optimizer.apply_gradients(zip(gradients, up))
    return ll, reg_penalty

@dataclass
class Model():
    gp: GP
    warp: Warp = None
    parameter_sample_size: Optional[int] = None
    locs: np.ndarray = None
    vals: np.ndarray = None
    cats: np.ndarray = None
    report: Callable = None
    verbose: bool = True

    def __post_init__(self):

        '''
        Parameters:
                x : Pandas DataFrame with columns for locations.

                u : A Pandas Series containing observations.

                featurization : function, optional
                    Should be a function that takes x1 (n-dim array of input data)
                    and returns the coordinates, i.e., x, y, x**2, y**2.
                    Example: def featurization(x1):
                                return x1[:, 0], x1[:, 1], x1[:, 0]**2, x1[:, 1]**2.
                    Default is None.

                latent : List[GP]
                     Name of the covariance function to use in the GP.
                     Should be 'squared-exp' or 'gamma-exp'.
                     Default is 'squared-exp'.

                verbose : boolean, optional
                    Whether or not to print parameters.
                    Default is True.

        Performs Gaussian process training and prediction.
        '''

        if self.warp is None: self.warp = NoWarp()

        # Default reporting function.
        def default_report(p, prefix=None):
            if prefix: print(prefix, end=' ')

            def fmt(x):
                if isinstance(x, tf.Tensor):
                    x = x.numpy()

                if isinstance(x, (int, np.int32, np.int64)):
                    return '{:5d}'.format(x)
                if isinstance(x, (float, np.float32, np.float64)):
                    return '{:5.2f}'.format(x)
                else:
                    with np.printoptions(precision=2, formatter={'floatkind': '{:5.2f}'.format}):
                        return str(x)

            print('[%s]' % (' '.join('%s %s' % (k, fmt(v)) for k, v in p.items())))

        if self.report == None: self.report = default_report

        if self.locs is not None: self.locs = np.array(self.locs)
        if self.vals is not None: self.vals = np.array(self.vals)
        if self.cats is not None: self.cats = np.array(self.cats)

        # Collect parameters and create TF parameters.
        for p in self.gather_vars().values():
            p.create_tf_variable()

    def gather_vars(self):
        return self.gp.gather_vars() | self.warp.gather_vars()

    def set(self, **values):
        parameters = self.gather_vars()
        for name, v in values.items():
            if name in parameters:
                parameters[name].value = v
                parameters[name].create_tf_variable()
            else:
                raise ValueError(f"{k} is not a parameter")
        return self

    def fit(self, locs, vals, cats=None, step_size=0.01, iters=100, reg=None):
        # Collect parameters and create TF parameters.
        parameters = self.gather_vars()

        # Permute datapoints if cats is given.
        if cats is not None:
            cats = np.array(cats)
            perm = np.argsort(cats)
            locs, vals, cats = locs[perm], vals[perm], cats[perm]
        else:
            cats = np.zeros(locs.shape[:1], np.int32)
            perm = None

        # Data dict.
        self.data = {
            'warplocs': self.warp(locs),
            'vals': tf.constant(vals, dtype=tf.float32),
            'cats': tf.constant(cats, dtype=tf.int32)}

        optimizer = tf.keras.optimizers.Adam(learning_rate=step_size)

        j = 0 # Iteration count.
        for i in range(10):
            t0 = time.time()
            while j < (i + 1) * iters / 10:
                ll, reg_penalty = gp_train_step(
                    optimizer, self.data, parameters, self.gp, reg)
                j += 1

            time_elapsed = time.time() - t0
            if self.verbose == True:
                self.report(
                  dict(iter=j, ll=ll, time=time_elapsed, reg=reg_penalty) |
                  {p.name: p.surface() for p in parameters.values()})

        # Save parameter values.
        for p in parameters.values():
            p.update_value()

        # Restore order if things were permuted.
        if perm is not None:
            revperm = np.argsort(perm)
            locs, vals, cats = locs[revperm], vals[revperm], cats[revperm]

        self.locs = locs
        self.vals = vals
        self.cats = cats

        return self

    def mcmc(self, locs, vals, cats=None,
            chains=4, step_size=0.1, move_prob=0.5,
            samples=1000, burnin=500, report_interval=100):

        assert samples % report_interval == 0, '`samples` must be a multiple of `report_interval`'
        assert burnin % report_interval == 0, '`burnin` must be a multiple of `report_interval`'

        # Permute datapoints if cats is given.
        if cats is not None:
            cats = np.array(cats)
            perm = np.argsort(cats)
            locs, vals, cats = locs[perm], vals[perm], cats[perm]

        # Data dict.
        self.data = {
            'locs': tf.constant(locs, dtype=tf.float32),
            'vals': tf.constant(vals, dtype=tf.float32),
            'cats': None if cats is None else tf.constant(cats, dtype=tf.int32)}

        # Initial MCMC state.
        initial_up = self.parameter_space.get_underlying(self.parameters)

        # Unnormalized log posterior distribution.
        def g(up):
            sp = self.parameter_space.get_surface(up)
            return gp_log_likelihood(self.data, sp, self.gp)

        def f(*up_flat):
            up = tf.nest.pack_sequence_as(initial_up, up_flat)
            ll = tf.map_fn(g, up, fn_output_signature=tf.float32)
            # log_prior = -tf.reduce_sum(tf.math.log(1. + tf.square(up_flat)), axis=0)
            return ll # + log_prior

        # Run the chain for a burst.
        @tf.function
        def run_chain(current_state, final_results, kernel, iters):
            samples, results, final_results = tfp.mcmc.sample_chain(
                num_results=iters,
                current_state=current_state,
                kernel=kernel,
                return_final_kernel_results=True,
                trace_fn=lambda _, results: results)

            return samples, results, final_results
        
        def new_state_fn(scale, dtype):
          direction_dist = tfd.Normal(loc=dtype(0), scale=dtype(1))
          scale_dist = tfd.Exponential(rate=dtype(1/scale))
          pick_dist = tfd.Bernoulli(probs=move_prob)

          def _fn(state_parts, seed):
            next_state_parts = []
            part_seeds = tfp.random.split_seed(
                seed, n=len(state_parts), salt='rwmcauchy')
            for sp, ps in zip(state_parts, part_seeds):
                pick = tf.cast(pick_dist.sample(sample_shape=sp.shape, seed=ps), tf.float32)
                direction = direction_dist.sample(sample_shape=sp.shape, seed=ps)
                scale_val = scale_dist.sample(seed=ps)
                next_state_parts.append(sp + tf.einsum('a...,a->a...', pick * direction, scale_val))
            return next_state_parts
          return _fn

        inv_temps = 0.5**np.arange(chains, dtype=np.float32)

        def make_kernel_fn(target_log_prob_fn):
            return tfp.mcmc.RandomWalkMetropolis(
                target_log_prob_fn=target_log_prob_fn,
                new_state_fn=new_state_fn(scale=step_size / np.sqrt(inv_temps), dtype=np.float32))

        kernel = tfp.mcmc.ReplicaExchangeMC(
            target_log_prob_fn=f,
            inverse_temperatures=inv_temps,
            make_kernel_fn=make_kernel_fn)

        # Do bursts.
        current_state = tf.nest.flatten(initial_up)
        final_results = None
        acc_states = []
        num_bursts = (samples + burnin) // report_interval
        burnin_bursts = burnin // report_interval
        for i in range(num_bursts):
            is_burnin = i < burnin_bursts

            if self.verbose and (i == 0 or i == burnin_bursts):
                print('BURNIN\n' if is_burnin else '\nSAMPLING')
            
            t0 = time.time()
            states, results, final_results = run_chain(current_state, final_results, kernel, report_interval)

            if self.verbose == True:
                if not is_burnin: print()
                accept_rates = results.post_swap_replica_results.is_accepted.numpy().mean(axis=0)
                print('[iter {:4d}] [time {:.1f}] [accept rates {}]'.format(
                    ((i if is_burnin else i - burnin_bursts) + 1) * report_interval,
                    time.time() - t0,
                    ' '.join([f'{x:.2f}' for x in accept_rates.tolist()])))

            if not is_burnin:
                acc_states.append(tf.nest.map_structure(lambda x: x.numpy(), states))
                all_states = [np.concatenate(x, 0) for x in zip(*acc_states)]
                up = tf.nest.pack_sequence_as(initial_up, all_states)
                sp = self.parameter_space.get_surface(up, numpy=True) 

                # Reporting
                if self.verbose == True:
                    for p in [5, 50, 95]:
                        x = tf.nest.map_structure(lambda x: np.percentile(x, p, axis=0), sp)
                        self.report(x, prefix=f'{p:02d}%ile')

            current_state = [s[-1] for s in states]

        posterior = self.parameter_space.get_surface(up, numpy=True)

        # Restore order if things were permuted.
        if cats is not None:
            revperm = np.argsort(perm)
            locs, vals, cats = locs[revperm], vals[revperm], cats[revperm]

        return replace(self, 
            parameters=posterior,
            parameter_sample_size=samples,
            locs=locs, vals=vals, cats=cats)

    def generate(self, locs, cats=None):
        assert self.locs is None and self.vals is None, 'Conditional generation not yet supported'
        assert self.parameter_sample_size is None, 'Generation from a distribution not yet supported'

        locs = np.array(locs)

        # Permute datapoints if cats is given.
        if cats is not None:
            cats = np.array(cats)
            perm = np.argsort(cats)
            locs, cats = locs[perm], cats[perm]
        else:
            cats = np.zeros(locs.shape[:1], np.int32)
            perm = None

        m, S = gp_covariance(
            self.gp,
            self.warp(locs).run({}),
            None if cats is None else tf.constant(cats, dtype=tf.int32))

        vals = MVN(m, tf.linalg.cholesky(S)).sample().numpy()

        # Restore order if things were permuted.
        if perm is not None:
            revperm = np.argsort(perm)
            locs, vals, cats = locs[revperm], vals[revperm], cats[revperm]

        self.locs = locs
        self.vals = vals
        self.cats = cats

        return self

    def predict(self, locs2, cats2=None, *, subsample=None, reduce=None, tracker=None, pair=False):
        '''
        Performs GP predictions of the mean and variance.
        Has support for batch predictions for large data sets.
        '''

        assert subsample is None or self.parameter_sample_size is not None, \
            '`subsample` is only valid with sampled parameters'

        assert reduce is None or self.parameter_sample_size is not None, \
            '`reduce` is only valid with sampled parameters'

        assert subsample is None or reduce is None, \
            '`subsample` and `reduce` cannot both be given'

        if tracker is None: tracker = lambda x: x

        assert self.locs.shape[-1] == locs2.shape[-1], 'Mismatch in location dimensions'
        if cats2 is not None:
            assert cats2.shape == locs2.shape[:1], 'Mismatched shapes in cats and locs'
        else:
            cats2 = np.zeros(locs2.shape[:1], np.int32)

        def interpolate_batch(A11i, locs1, vals1diff, cats1, locs2, cats2):
            """
            Inputs:
              locs1.shape = [N1, K]
              vals1diff.shape = [N1]
              cats1.shape = [N1]
              locs2.shape = [N2, K]
              cats2.shape = [N2]

            Outputs:
              u2_mean.shape = [N2]
              u2_var.shape = [N2]
            """

            N1 = len(locs1) # Number of measurements.

            # Permute datapoints if cats is given.
            if cats2 is not None:
                perm = np.argsort(cats2)
                locs2, cats2 = locs2[perm], cats2[perm]
                locs2 = self.warp(locs2).run({})

            _, A12 = gp_covariance2(
                self.gp,
                tf.constant(locs1, dtype=tf.float32),
                tf.constant(cats1, dtype=tf.int32),
                tf.constant(locs2, dtype=tf.float32),
                tf.constant(cats2, dtype=tf.int32),
                N1)

            m2, A22 = gp_covariance(
                self.gp,
                tf.constant(locs2, dtype=tf.float32),
                tf.constant(cats2, dtype=tf.int32))

            # Restore order if things were permuted.
            if cats2 is not None:
                revperm = np.argsort(perm)
                m2 = tf.gather(m2, revperm)
                A12 = tf.gather(A12, revperm, axis=-1)
                A22 = tf.gather(tf.gather(A22, revperm), revperm, axis=-1)

            u2_mean = m2 + tf.einsum('ab,a->b', A12, tf.einsum('ab,b->a', A11i, vals1diff))
            u2_var = tf.linalg.diag_part(A22) -  tf.einsum('ab,ab->b', A12, tf.matmul(A11i, A12))

            return u2_mean, u2_var

        def interpolate_pair_batch(A11i, locs1, vals1diff, cats1, locs2, cats2):
            """
            Inputs:
              locs1.shape = [N1, K]
              vals1diff.shape = [N1]
              cats1.shape = [N1]
              locs2.shape = [N2, 2, K]
              cats2.shape = [N2]

            Outputs:
              u2_mean.shape = [N2, 2]
              u2_var.shape = [N2, 2, 2]
            """

            N1 = len(locs1) # Number of measurements.
            N2 = len(locs2) # Number of prediction pairs.

            # Permute datapoints if cats is given.
            perm = np.argsort(cats2)
            locs2, cats2 = locs2[perm], cats2[perm]

            # Warp locs2.
            locs2_shape = locs2.shape
            locs2 = locs2.reshape([-1, locs2_shape[-1]])  # Shape into matrix.
            locs2 = self.warp(locs2).run({})
            locs2 = tf.reshape(locs2, locs2_shape)  # Revert shape.

            _, A12 = gp_covariance2(
                self.gp,
                tf.constant(locs1, dtype=tf.float32),
                tf.constant(cats1, dtype=tf.int32),
                tf.constant(locs2[:, 0, :], dtype=tf.float32),
                tf.constant(cats2, dtype=tf.int32),
                N1)

            _, A13 = gp_covariance2(
                self.gp,
                tf.constant(locs1, dtype=tf.float32),
                tf.constant(cats1, dtype=tf.int32),
                tf.constant(locs2[:, 1, :], dtype=tf.float32),
                tf.constant(cats2, dtype=tf.int32),
                N1)

            m2, A22 = gp_covariance(
                self.gp,
                tf.constant(locs2[:, 0, :], dtype=tf.float32),
                tf.constant(cats2, dtype=tf.int32))

            m3, A33 = gp_covariance(
                self.gp,
                tf.constant(locs2[:, 1, :], dtype=tf.float32),
                tf.constant(cats2, dtype=tf.int32))

            _, A23 = gp_covariance2(
                self.gp,
                tf.constant(locs2[:, 0, :], dtype=tf.float32),
                tf.constant(cats2, dtype=tf.int32),
                tf.constant(locs2[:, 1, :], dtype=tf.float32),
                tf.constant(cats2, dtype=tf.int32),
                N2)

            # Reassemble into more useful shapes.

            A12 = tf.stack([A12, A13], axis=-1) # [N1, N2, 2]

            m2 = tf.stack([m2, m3], axis=-1) # [N2, 2]

            A22 = tf.linalg.diag_part(A22)
            A33 = tf.linalg.diag_part(A33)
            A23 = tf.linalg.diag_part(A23)
            A22 = tf.stack([tf.stack([A22, A23], axis=-1), tf.stack([A23, A33], axis=-1)], axis=-2) # [N2, 2, 2]

            # Restore order if things were permuted.
            revperm = np.argsort(perm)
            m2 = tf.gather(m2, revperm)
            A12 = tf.gather(A12, revperm, axis=1)
            A22 = tf.gather(A22, revperm)

            u2_mean = m2 + tf.einsum('abc,a->bc', A12, tf.einsum('ab,b->a', A11i, vals1diff))
            u2_var = A22 - tf.einsum('abc,abd->bcd', A12, tf.einsum('ae,ebd->abd', A11i, A12))

            return u2_mean, u2_var

        def interpolate(locs1, vals1, cats1, locs2, cats2, pair=False):
            # Interpolate in batches.
            batch_size = locs1.shape[0] // 2

            for_gp = []

            for start in np.arange(0, len(locs2), batch_size):
                stop = start + batch_size
                subset = locs2[start:stop], cats2[start:stop]
                for_gp.append(subset)

            # Permute datapoints if cats is given.
            if cats1 is not None:
                perm = np.argsort(cats1)
                locs1, vals1, cats1 = locs1[perm], vals1[perm], cats1[perm]

            locs1 = self.warp(locs1).run({})

            m1, A11 = gp_covariance(
                self.gp,
                tf.constant(locs1, dtype=tf.float32),
                tf.constant(cats1, dtype=tf.int32))

            A11i = tf.linalg.inv(A11)

            u2_mean_s = []
            u2_var_s = []

            f = interpolate_pair_batch if pair else interpolate_batch

            for locs_subset, cats_subset in for_gp:
                u2_mean, u2_var = f(A11i, locs1, vals1 - m1, cats1, locs_subset, cats_subset)
                u2_mean = u2_mean.numpy()
                u2_var = u2_var.numpy()
                u2_mean_s.append(u2_mean)
                u2_var_s.append(u2_var)

            u2_mean = np.concatenate(u2_mean_s)
            u2_var = np.concatenate(u2_var_s)

            return u2_mean, u2_var

        if self.parameter_sample_size is None:
            m, v = interpolate(self.locs, self.vals, self.cats, locs2, cats2, pair)
        elif reduce == 'median':
            raise NotImplementedError
            p = tf.nest.map_structure(lambda x: np.quantile(x, 0.5, axis=0).astype(np.float32), self.parameters)
            m, v = interpolate(self.locs, self.vals, self.cats, locs2, cats2, p, pair)
        elif reduce == 'mean':
            raise NotImplementedError
            p = tf.nest.map_structure(lambda x: x.mean(axis=0).astype(np.float32), self.parameters)
            m, v = interpolate(self.locs, self.vals, self.cats, locs2, cats2, p, pair)
        else:
            raise NotImplementedError
            samples = self.parameter_sample_size

            if subsample is not None:
                assert subsample <= samples, '`subsample` may not exceed sample size'
            else:
                subsample = samples

            # Thin by picking roughly equally-spaced samples.
            a = np.arange(samples) * subsample / samples % 1
            pick = np.concatenate([[True], a[1:] >= a[:-1]])
            parameters = tf.nest.map_structure(lambda x: x[pick], self.parameters)

            # Make a prediction for each sample.
            results = []
            for i in tracker(range(subsample)):
                p = tf.nest.map_structure(lambda x: x[i], parameters)
                results.append(interpolate(self.locs, self.vals, self.cats, locs2, cats2, p, pair))

            mm, vv = [np.stack(x) for x in zip(*results)]
            m = mm.mean(axis=0)
            v = (np.square(mm) + vv).mean(axis=0) - np.square(m)

        return m, v



from dataclasses import dataclass
from typing import Callable, List, Union
import numpy as np

# Tensorflow is extraordinarily noisy. Catch warnings during import.
import warnings
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    import tensorflow as tf
    from tensorflow.linalg import LinearOperatorFullMatrix as LOFullMatrix
    from tensorflow.linalg import LinearOperatorBlockDiag as LOBlockDiag

from .op import Op
from .metric import Euclidean, PerAxisDist2, ed
from .param import ppp, upp, bpp, ppp_list
from .mean import get_trend_coefs

__all__ = ['Kernel']

def block_diag(blocks):
    """Return a dense block-diagonal matrix."""
    return LOBlockDiag([LOFullMatrix(b) for b in blocks]).to_dense()

class Kernel(Op):
    def __init__(self, fa, autoinputs):
        if 'offset' not in autoinputs: autoinputs['offset'] = 'offset'
        if 'locs1' not in autoinputs: autoinputs['locs1'] = 'locs1'
        if 'locs2' not in autoinputs: autoinputs['locs2'] = 'locs2'
        super().__init__(fa, autoinputs)

    def __add__(self, other):
        if other is None:
            return self
        else:
            return Stack([self]) + other

    def __mul__(self, other):
        return Product([self]) * other

    def call(self, e):
        """
        Returns tuple `(mean, covariance)` for locations.
        Return values may be unbroadcasted.
        """
        pass

    def __call__(self, e):
        """
        Returns tuple `(mean, covariance)` for locations.
        Return values have correct shapes.
        """
        C = self.call(e)
        if C is None: C = 0.
        n1 = tf.shape(e['locs1'])[0]
        n2 = tf.shape(e['locs2'])[0]
        C = tf.broadcast_to(C, [n1, n2])
        return C

    def report(self):
        string = ', '.join('%s %4.2f' % (v.name, p[v.name]) for v in self.vars())
        return '[' + string + ']'

class TrendPrior(Kernel):
    def __init__(self, featurizer, alpha):
        fa = dict(alpha=alpha)
        self.featurizer = featurizer
        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['alpha'])

    def call(self, e):
        F1 = tf.cast(self.featurizer(e['locs1']), tf.float32)
        F2 = tf.cast(self.featurizer(e['locs2']), tf.float32)
        return e['alpha'] * tf.einsum('ba,ca->bc', F1, F2)

def scale_to_metric(scale, metric):
    assert scale is None or metric is None
    if metric is None:
        if scale is None:
            metric = 'euclidean'
        else:
            metric = Euclidean(scale)
    return metric

class Constant(Kernel):
    def __init__(self, sill):
        fa = dict(sill=sill)
        super().__init__(fa, dict())

    def vars(self):
        return ppp(self.fa['sill'])

    def call(self, e):
        return e['sill']

class SquaredExponential(Kernel):
    def __init__(self, sill, range, scale=None, metric=None):
        fa = dict(sill=sill, range=range)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp(self.fa['sill']) | ppp(self.fa['range'])

    def call(self, e):
        return e['sill'] * tf.exp(-0.5 * e['d2'] / tf.square(e['range']))

class GammaExponential(Kernel):
    def __init__(self, range, sill, gamma, scale=None, metric=None):
        fa = dict(sill=sill, range=range, gamma=gamma, scale=scale)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp(self.fa['sill']) | ppp(self.fa['range']) | bpp(self.fa['gamma'], 0., 2.)

    def call(self, e):
        return e['sill'] * gamma_exp(e['d2'] / tf.square(e['range']), e['gamma'])

@tf.custom_gradient
def ramp(x):
    ax = tf.abs(x)
    def grad(upstream):
        return upstream * tf.where(ax < 1., -tf.sign(x), 0.)
    return tf.maximum(0., 1. - ax), grad

class Ramp(Kernel):
    def __init__(self, range, sill, scale=None, metric=None):
        fa = dict(sill=sill, range=range, scale=scale)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp(self.fa['sill']) | ppp(self.fa['range'])

    def call(self, e):
        return e['sill'] * ramp(tf.sqrt(e['d2']) / e['range'])

# @tf.custom_gradient
# def rampstack(x, sills, ranges):
#     """
#     `x` has arbitrary shape [...].
#     `sills` and `ranges` both have shape [K].
#     """
#     ax = ed(tf.abs(x)) # [..., 1]
#     y = sills * tf.maximum(0., 1. - ax / ranges) # [..., K]
#     def grad(upstream):
#         ax = ed(tf.abs(x)) # [..., 1]
#         y = sills * tf.maximum(0., 1. - ax / ranges) # [..., K]
#         K = tf.shape(sills)[0]
#         grad_x = upstream * tf.reduce_sum(tf.where(ax < ranges, -tf.sign(ed(x)), 0.), -1) # [...]
#         grad_sills = tf.reduce_sum(tf.reshape(ed(upstream) * y, [-1, K]), 0) # [K}
#         grad_ranges = tf.where(ax < ranges, sills * ax / tf.square(ranges), 0.) # [..., K}
#         grad_ranges = tf.reduce_sum(tf.reshape(ed(upstream) * grad_ranges, [-1, K]), 0) # [K]
#         return grad_x, grad_sills, grad_ranges
#     return tf.reduce_sum(y, -1), grad

@tf.custom_gradient
def rampstack(x, sills, ranges):
    """
    `x` has arbitrary shape [...], but must be non-negative.
    `sills` and `ranges` both have shape [K].
    """
    ax = ed(tf.abs(x)) # [..., 1]
    y = sills * tf.maximum(0., 1. - ax / ranges) # [..., K]
    def grad(upstream):
        ax = ed(tf.abs(x)) # [..., 1]
        y = sills * tf.maximum(0., 1. - ax / ranges) # [..., K]
        K = tf.shape(sills)[0]
        small = ax < ranges
        grad_x = upstream * tf.reduce_sum(tf.where(small, -tf.sign(ed(x)) * (sills / ranges), 0.), -1) # [...]
        grad_sills = tf.einsum('ak,a->k', tf.reshape(y, [-1, K]), tf.reshape(upstream, [-1]))
        grad_ranges = tf.where(small, ax * (sills / tf.square(ranges)), 0.) # [..., K}
        grad_ranges = tf.einsum('ak,a->k', tf.reshape(grad_ranges, [-1, K]), tf.reshape(upstream, [-1]))
        return grad_x, grad_sills, grad_ranges
    return tf.reduce_sum(y, -1), grad

class RampStack(Kernel):
    def __init__(self, range, sill, scale=None, metric=None):
        fa = dict(sill=sill, range=range, scale=scale)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp_list(self.fa['sill']) | ppp_list(self.fa['range'])

    def call(self, e):
        if isinstance(e['sill'], (tuple, list)):
            e['sill'] = tf.stack(e['sill'])
        if isinstance(e['range'], (tuple, list)):
            e['range'] = tf.stack(e['range'])

        return rampstack(tf.sqrt(e['d2']), e['sill'], e['range'])

@tf.recompute_grad
def smooth_convex(x, sills, ranges):
    """
    `x` has arbitrary shape [...], but must be non-negative.
    `sills` and `ranges` both have shape [K].
    """
    r2 = ranges
    r1 = tf.pad(ranges[:-1], [[1, 0]])
    ex = ed(x)
    ax = tf.abs(ex)
    rx = ax / r2 - 1.

    c1 = 2. / (r1 + r2)
    c2 = 1. / (1. - tf.square(r1/r2))

    # i1 = tf.cast(ax <= r1, tf.float32) # Indicates x <= r1.
    # i2 = tf.cast(ax <= r2, tf.float32) * (1. - i1) # Indicates r1 < x <= r2.
    # v = i1 * (1. - c1 * ax) + i2 * c2 * tf.square(rx)

    v = tf.where(ax <= r1, 1. - c1 * ax, c2 * tf.square(rx))
    v = tf.where(ax <= r2, v, 0.)

    y = tf.einsum('...k,k->...', v, sills)
    return y

@tf.custom_gradient
def smooth_convex_grad(x, sills, ranges):
    """
    `x` has arbitrary shape [...], but must be non-negative.
    `sills` and `ranges` both have shape [K].
    """
    r2 = ranges
    r1 = tf.pad(ranges[:-1], [[1, 0]])
    ex = ed(x)
    ax = tf.abs(ex)
    rx = ax / r2 - 1.

    c1 = 2. / (r1 + r2)
    c2 = 1. / (1. - tf.square(r1/r2))

    v = tf.where(ax <= r1, 1. - c1 * ax, c2 * tf.square(rx))
    v = tf.where(ax <= r2, v, 0.)

    y = tf.einsum('...k,k->...', v, sills)

    def grad(upstream):
        r2 = ranges
        r1 = tf.pad(ranges[:-1], [[1, 0]])
        ex = ed(x)
        ax = tf.abs(ex)
        rx = ax / r2 - 1.
        i1 = tf.cast(ax <= r1, tf.float32) # Indicates x <= r1.
        i2 = tf.cast(ax <= r2, tf.float32) * (1. - i1) # Indicates r1 < x <= r2.

        c1 = 2. / (r1 + r2)
        c2 = 1. / (1. - tf.square(r1 / r2))
        c3 = 1. / (r2 - tf.square(r1) / r2)

        v = i1 * (1. - c1 * ax) + i2 * c2 * tf.square(rx)

        sx = tf.sign(ex)

        K = tf.shape(sills)[0]
        gx = sx * sills * (i1 * -c1 + i2 * rx * (2 * c3))
        grad_x = upstream * tf.reduce_sum(gx, -1) # [...]

        grad_sills = tf.einsum('ak,a->k', tf.reshape(v, [-1, K]), tf.reshape(upstream, [-1]))

        u = 2 / tf.square(r1 + r2) * ax * i1
        yr1 = u + i2 * tf.square(rx * c3) * 2 * r1
        yr2 = u - 2 * i2 * (rx * c3 + tf.square(rx * c2) / r2)
        yr1 = sills * tf.reshape(yr1, [-1, K])
        yr2 = sills * tf.reshape(yr2, [-1, K])
        yr = tf.pad(yr1[:, 1:], [[0, 0], [0, 1]]) + yr2
        grad_ranges = tf.einsum('ak,a->k', yr, tf.reshape(upstream, [-1]))

        return grad_x, grad_sills, grad_ranges

    return y, grad

class SmoothConvex(Kernel):
    def __init__(self, range, sill, scale=None, metric=None):
        fa = dict(sill=sill, range=range, scale=scale)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp_list(self.fa['sill']) | ppp_list(self.fa['range'])

    def call(self, e):
        if isinstance(e['sill'], (tuple, list)):
            e['sill'] = tf.stack(e['sill'])
        if isinstance(e['range'], (tuple, list)):
            e['range'] = tf.stack(e['range'])

        return smooth_convex(tf.sqrt(e['d2']), e['sill'], e['range'])

@tf.recompute_grad
def quadstack(x, sills, ranges):
    """
    `x` has arbitrary shape [...], but must be non-negative.
    `sills` and `ranges` both have shape [K].
    """
    ex = ed(x)
    ax = tf.maximum(0., 1. - tf.abs(ex) / ranges) # [..., 1]
    y = sills * tf.square(ax) # [..., K]
    return tf.reduce_sum(y, -1)

class QuadStack(Kernel):
    def __init__(self, range, sill, scale=None, metric=None):
        fa = dict(sill=sill, range=range, scale=scale)
        autoinputs = scale_to_metric(scale, metric)
        super().__init__(fa, dict(d2=autoinputs))

    def vars(self):
        return ppp_list(self.fa['sill']) | ppp_list(self.fa['range'])

    def call(self, e):
        if isinstance(e['sill'], (tuple, list)):
            e['sill'] = tf.stack(e['sill'])
        if isinstance(e['range'], (tuple, list)):
            e['range'] = tf.stack(e['range'])

        return quadstack(tf.sqrt(e['d2']), e['sill'], e['range'])

class Wiener(Kernel):
    def __init__(self, axis, start):

        self.axis = axis
        self.start = start

        # Include the element of scale corresponding to the axis of
        # integration as an explicit formal argument.
        fa = dict()

        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return {}

    def call(self, e):
        x1 = e['locs1'][..., self.axis]
        x2 = e['locs2'][..., self.axis]
        k = tf.maximum(0., tf.minimum(ed(x1, 1), ed(x2, 0)) - self.start)
        return k

class IntSquaredExponential(Kernel):
    def __init__(self, axis, start, range):

        self.axis = axis
        self.start = start

        # Include the element of scale corresponding to the axis of
        # integration as an explicit formal argument.
        fa = dict(range=range)

        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['range'])

    def call(self, e):
        x1 = tf.pad(e['locs1'][..., self.axis] - self.start, [[1, 0]])
        x2 = tf.pad(e['locs2'][..., self.axis] - self.start, [[1, 0]])

        r = e['range']
        sdiff = (ed(x1, 1) - ed(x2, 0)) / (r * np.sqrt(2.))
        k = -tf.square(r) * (np.sqrt(np.pi) * sdiff * tf.math.erf(sdiff) + tf.exp(-tf.square(sdiff)))
        k -= k[0:1, :]
        k -= k[:, 0:1]
        k = k[1:, 1:]
        k = tf.maximum(0., k)

        return k

class IntExponential(Kernel):
    def __init__(self, axis, start, range):

        self.axis = axis
        self.start = start

        # Include the element of scale corresponding to the axis of
        # integration as an explicit formal argument.
        fa = dict(range=range)

        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['range'])

    def call(self, e):
        x1 = tf.pad(e['locs1'][..., self.axis] - self.start, [[1, 0]])
        x2 = tf.pad(e['locs2'][..., self.axis] - self.start, [[1, 0]])

        r = e['range']
        sdiff = tf.abs(ed(x1, 1) - ed(x2, 0)) / r
        k = -tf.square(r) * (sdiff + tf.exp(-sdiff))
        k -= k[0:1, :]
        k -= k[:, 0:1]
        k = k[1:, 1:]
        k = tf.maximum(0., k)

        return k

class Noise(Kernel):
    def __init__(self, nugget):
        fa = dict(nugget=nugget)
        super().__init__(fa, dict(locs1='locs1', locs2='locs2'))

    def vars(self):
        return ppp(self.fa['nugget'])

    def call(self, e):

        indices1 = tf.range(tf.shape(e['locs1'])[0])
        indices2 = tf.range(tf.shape(e['locs2'])[0]) + e['offset']
        C = tf.where(tf.equal(tf.expand_dims(indices1, -1), indices2), e['nugget'], 0.)
        return C

class Delta(Kernel):
    def __init__(self, sill, axes=None):
        fa = dict(sill=sill)
        self.axes = axes
        super().__init__(fa, dict(pa_d2='per_axis_dist2'))

    def vars(self):
        return ppp(self.fa['sill'])

    def call(self, e):

        if self.axes is not None:
            n = tf.shape(e['pa_d2'])[-1]
            mask = tf.math.bincount(self.axes, minlength=n, maxlength=n, dtype=tf.float32)
            d2 = tf.einsum('abc,c->ab', e['pa_d2'], mask)
        else:
            d2 = tf.reduce_sum(e['pa_d2'], axis=-1)

        return e['sill'] * tf.cast(tf.equal(d2, 0.), tf.float32)

class Mix(Kernel):
    def __init__(self, inputs, weights=None):
        self.inputs = inputs
        fa = {}
        ai = dict(cats1='cats1', cats2='cats2')

        # Special case if weights is not given.
        if weights is not None:
            fa['weights'] = weights
            ai['inputs'] = inputs

        super().__init__(fa, ai)

    def gather_vars(self, cache=None):
        """Make a special version of gather_vars because
           we want to gather variables from `inputs`
           even when it's not in autoinputs"""
        vv = super().gather_vars(cache)
        for iput in self.inputs:
            cache[id(self)] |= iput.gather_vars(cache)
        return cache[id(self)]

    def vars(self):
        if 'weights' in self.fa:
            return {k: p for row in self.fa['weights']
                      for k, p in get_trend_coefs(row).items()}
        else:
            return {}

    def call(self, e):
        if 'weights' in e:
            weights = []
            for row in e['weights']:
                if isinstance(row, (tuple, list)):
                    row = tf.stack(row)
                    weights.append(row)
            weights = tf.stack(weights)
            C = tf.stack(e['inputs'], axis=-1) # [locs, locs, numinputs].
            Aaug1 = tf.gather(weights, e['cats1']) # [locs, numinputs].
            Aaug2 = tf.gather(weights, e['cats2']) # [locs, numinputs].
            outer = tf.einsum('ac,bc->abc', Aaug1, Aaug2) # [locs, locs, numinputs].
            C = tf.einsum('abc,abc->ab', C, outer) # [locs, locs].
            return C
        else:
            # When weights is not given, exploit the fact that we don't have
            # to compute every element in component covariance matrices.
            N = len(self.inputs)
            catcounts1 = tf.math.bincount(e['cats1'], minlength=N, maxlength=N)
            catcounts2 = tf.math.bincount(e['cats2'], minlength=N, maxlength=N)
            catindices1 = tf.math.cumsum(catcounts1, exclusive=True)
            catindices2 = tf.math.cumsum(catcounts2, exclusive=True)
            catdiffs = tf.unstack(catindices2 - catindices1, num=N)
            locsegs1 = tf.split(e['locs1'], catcounts1, num=N)
            locsegs2 = tf.split(e['locs2'], catcounts2, num=N)

            # TODO: Check that the below is still correct.
            CC = [] # Observation noise submatrices.
            for sublocs1, sublocs2, catdiff, iput in zip(locsegs1, locsegs2, catdiffs, self.inputs):
                cache = dict(
                    offset = e['offset'] + catdiff,
                    locs1 = sublocs1,
                    locs2 = sublocs2)
                cache['per_axis_dist2'] = PerAxisDist2().run(cache)
                cache['euclidean'] = Euclidean().run(cache)
                Csub = iput.run(cache)
                CC.append(Csub)

            return block_diag(CC)

class Stack(Kernel):
    def __init__(self, parts: List[Kernel]):
        self.parts = parts
        super().__init__({}, dict(locs1='locs1', locs2='locs2', parts=parts))

    def vars(self):
        return {k: p for part in self.parts for k, p in part.vars().items()}

    def __add__(self, other):
        if isinstance(other, Kernel):
            return Stack(self.parts + [other])
    
    def call(self, e):
        return tf.reduce_sum(e['parts'], axis=0)

    def report(self):
        return ' '.join(part.report(p) for part in self.parts)

class Product(Kernel):
    def __init__(self, parts: List[Kernel]):
        self.parts = parts
        super().__init__({}, dict(locs1='locs1', locs2='locs2', parts=parts))

    def vars(self):
        return {k: p for part in self.parts for k, p in part.vars().items()}

    def __mul__(self, other):
        if isinstance(other, Kernel):
            return Product(self.parts + [other])
    
    def call(self, e):
        return tf.reduce_prod(e['parts'], axis=0)

    def report(self):
        return ' '.join(part.report(p) for part in self.parts)

# Gamma exponential covariance function.
@tf.custom_gradient
def safepow(x, a):
    y = tf.pow(x, a)
    def grad(dy):
        dx = tf.where(x <= 0.0, tf.zeros_like(x), dy * tf.pow(x, a-1))
        dx = unbroadcast(dx, x.shape)
        da = tf.where(x <= 0.0, tf.zeros_like(a), dy * y * tf.math.log(x))
        da = unbroadcast(da, a.shape)
        return dx, da
    return y, grad

def unbroadcast(x, shape):
    xrank = len(x.shape)
    rank = len(shape)
    excess_rank = tf.maximum(0, xrank - rank)
    x = tf.reduce_sum(x, axis=tf.range(excess_rank))
    axes_that_are_one = tf.where(tf.equal(shape, 1))[:, 0]
    x = tf.reduce_sum(x, axis=axes_that_are_one, keepdims=True)
    return x

def gamma_exp(d2, gamma):
    return tf.exp(-safepow(tf.maximum(d2, 0.0), 0.5 * gamma))

class Observation(Op):

    def __init__(self,
        coefs: List,
        noise: Kernel
    ):
        self.coefs = coefs
        self.noise = noise
        super().__init__({}, self.noise)

    def vars(self):
        vv = {k: p for c in self.coefs for k, p in upp(c)}
        vv |= self.noise.vars()
        return vv

    def __call__(self, e):
        """
        Dummy.
        """
        return 0.



import numpy as np
from geostat import Mesh
from shapely.geometry import Polygon

def test_mesh_from_convex_hull():

    # Create 100 random locations in a square centered on the origin.
    locs = np.random.uniform(-1., 1., [100, 2])

    mesh = Mesh.from_convex_hull(locs, 20)

    values = np.random.uniform(-10., 10., [mesh.locations().shape[0]])

    x, y, u = mesh.slice(values)

    print(np.sum(np.isnan(u)))

    assert x.shape == y.shape == u.shape

def test_mesh_from_polygon():

    polygon = Polygon([(0., 0.), (1., 1.), (1., 0.)])

    mesh = Mesh.from_polygon(polygon, 20)

    values = np.random.uniform(-10., 10., [mesh.locations().shape[0]])

    x, y, u = mesh.slice(values)

    print(np.sum(np.isnan(u)))

    assert x.shape == y.shape == u.shape



import numpy as np
import tensorflow as tf
import geostat
import geostat.kernel as krn
from geostat import Parameters, GP, Model, Trend

def test_noise():
    np.random.seed(2)
    tf.random.set_seed(2)

    # Create random locations in a square centered on the origin.
    locs1 = np.random.normal(size=[1000, 2])

    # Create parameters.
    p = Parameters(nugget=1.)

    # Create model and generate data.
    kernel = krn.Noise(p.nugget)
    model = Model(GP(0, kernel))
    vals1 = model.generate(locs1).vals

    # Fit GP.
    model.set(nugget=0.5)
    model.fit(locs1, vals1, iters=50, step_size=1e-1)

    assert np.allclose(
        [getattr(p, name).value for name in ['nugget']],
        [1.],
        rtol=0.3)

    # Interpolate using GP.
    N = 20
    xx, yy = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy], axis=-1).reshape([-1, 2])

    mean, var = model.predict(locs2)
    mean2, var2 = model.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)

def test_gp_with_trend():
    np.random.seed(2)
    tf.random.set_seed(2)

    # Create random locations in a square centered on the origin.
    locs1 = np.random.normal(size=[1000, 2])

    # Create parameters.
    p = Parameters(range=0.33, nugget=1., beta=[4., 3., 2., 1.])

    # Featurizer of location for trends.
    @geostat.featurizer()
    def trend_featurizer(x, y): return 1., x, y, x*y

    # Define model.
    trend = Trend(trend_featurizer, beta=p.beta)
    kernel = krn.SquaredExponential(sill=1., range=p.range) + krn.Noise(nugget=p.nugget)
    model = Model(GP(trend, kernel))

    # Generate data.
    vals1 = model.generate(locs1).vals

    # Fit GP.
    model.set(range=1, nugget=2., beta=[1., 2., 3., 4.])
    model.fit(locs1, vals1, iters=100, step_size=1e-1)

    assert np.allclose(p.beta.value, [4., 3., 2., 1.], rtol=0.3)

    assert np.allclose(
        [getattr(p, name).value for name in ['range', 'nugget']],
        [0.33, 1.],
        rtol=0.3)

    # Interpolate using GP.
    N = 20
    xx, yy = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy], axis=-1).reshape([-1, 2])

    mean, var = model.predict(locs2)
    mean2, var2 = model.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)

def test_gp2d():
    np.random.seed(2)
    tf.random.set_seed(2)

    # Create random locations in a square centered on the origin.
    locs1 = np.random.normal(size=[1000, 2])

    # Create parameters.
    p = Parameters(alpha=1., range=0.33, nugget=1.)

    # Featurizer of location for trends.
    @geostat.featurizer(normalize=locs1)
    def trend_featurizer(x, y): return x, y, x*y

    # Make model.
    kernel = krn.TrendPrior(trend_featurizer, alpha=p.alpha) \
           + krn.SquaredExponential(sill=1., range=p.range) \
           + krn.Noise(p.nugget)
    model = Model(GP(0, kernel))

    # Generate data.
    vals1 = model.generate(locs1).vals

    # Fit GP.
    model.set(alpha=2., range=1., nugget=0.5)
    model.fit(locs1, vals1, iters=100, step_size=1e-1)

    assert np.allclose(
        [getattr(p, name).value for name in ['range', 'nugget']],
        [0.33, 1.],
        rtol=0.3)

    # Interpolate using GP.
    N = 20
    xx, yy = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy], axis=-1).reshape([-1, 2])

    mean, var = model.predict(locs2)
    mean2, var2 = model.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)

def test_gp3d():
    np.random.seed(2)
    tf.random.set_seed(2)

    # Create random locations in a square centered on the origin.
    locs1 = np.random.normal(size=[600, 3])
    locs1 = np.concatenate([locs1, locs1 * [1., 1., 0.8], locs1 * [1., 1., 1.1]])

    # Create parameters.
    p = Parameters(alpha=1., zscale=5., range=0.5, sill=1., gamma=1., dsill=0.1, nugget=0.1)

    # Featurizer of location for trends.
    @geostat.featurizer(normalize=locs1)
    def trend_featurizer(x, y, z): return z, z*z

    # Create model.
    kernel = \
        krn.TrendPrior(trend_featurizer, alpha=p.alpha) + \
        krn.GammaExponential(range=p.range, sill=p.sill, gamma=p.gamma, scale=[1., 1., p.zscale]) + \
        krn.Delta(axes=[0, 1], sill=p.dsill) + \
        krn.Noise(nugget=p.nugget)
    model = Model(GP(0, kernel))

    # Generate data.
    vals1 = model.generate(locs1).vals

    # Fit GP.
    model.set(alpha=2., zscale=1., range=1., sill=0.5, gamma=0.5, dsill=0.5, nugget=0.5)
    model.fit(locs1, vals1, iters=200, step_size=1e-1)

    assert np.allclose(
        [getattr(p, name).value for name in ['zscale', 'range', 'sill', 'gamma', 'dsill', 'nugget']],
        [5., 0.5, 1., 1., 0.1, 0.1],
        rtol=0.5)

    # Interpolate using GP.
    N = 10
    xx, yy, zz = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy, zz], axis=-1).reshape([-1, 3])

    mean, var = model.predict(locs2)
    mean2, var2 = model.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)

def test_gp3d_stacked():
    np.random.seed(2)
    tf.random.set_seed(2)

    # Create random locations centered on the origin.
    locs1 = np.random.normal(size=[2500, 3])

    # Create parameters.
    p = Parameters(alpha=1., zscale=5., r1=0.25, s1=1., r2=1.0, s2=0.25, nugget=1.)

    # Featurizer of location for trends.
    @geostat.featurizer(normalize=locs1)
    def trend_featurizer(x, y, z): return z, z*z

    # Covariance structure
    kernel = \
        krn.TrendPrior(trend_featurizer, alpha=p.alpha) + \
        krn.SquaredExponential(range=p.r1, sill=p.s1, scale=[1., 1., p.zscale]) + \
        krn.SquaredExponential(range=p.r2, sill=p.s2, scale=[1., 1., 0.]) + \
        krn.Noise(nugget=p.nugget)

    # Model
    model = Model(GP(0, kernel))

    # Generate data.
    vals1 = model.generate(locs1).vals

    # Fit GP.
    model.set(alpha=2., zscale=2.5, r1=0.125, s1=0.5, r2=0.5, s2=0.125, nugget=0.5)
    model.fit(locs1, vals1, iters=100, step_size=1e-1)

    assert np.allclose(
        [getattr(p, name).value for name in ['zscale', 'r1', 's1', 'r2', 's2', 'nugget']],
        [5., 0.25, 1., 1., 0.25, 1.],
        rtol=0.3)

    # Interpolate using GP.
    N = 10
    xx, yy, zz = np.meshgrid(np.linspace(-1, 1, N), np.linspace(-1, 1, N), np.linspace(-1, 1, N))
    locs2 = np.stack([xx, yy, zz], axis=-1).reshape([-1, 3])

    mean, var = model.predict(locs2)
    mean2, var2 = model.predict(locs2)

    assert np.all(mean == mean2)
    assert np.all(var == var2)



from .op import Op
from .param import get_parameter_values


__all__ = ['function']


class CustomOp(Op):
    def __init__(self, f, **fa):
        self.f = f
        super().__init__(fa, {})
    def __call__(self, e):
        return self.f(*e.values())
    def vars(self):
        # Assume for now that all arguments are Parameters
        return {p.name: p for p in fa.values()}


def function(f):
    """
    A custom op involves two functions:
      * f, where it is defined, and
      * g, where it is introduced into the graph.
    """
    def g(*args):
      # Assume for now that all arguments are Parameters
      return CustomOp(f, **{p.name: p for p in args})
    return g
